package kieker.webgui.common;

import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import javax.annotation.PostConstruct;

import kieker.analysis.model.analysisMetaModel.MIDisplayConnector;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.MIView;
import kieker.analysis.plugin.AbstractPlugin;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.common.exception.AnalysisAlreadyStartedException;
import kieker.webgui.common.exception.AnalysisInstantiationException;
import kieker.webgui.common.exception.ProjectLoadException;

// TODO AnalysisController, Thread etc. have to be initialized via the class loader as well
public class Analysis {
	/**
	 * This is the log for errors, exceptions etc.
	 */
	private static final Log LOG = LogFactory.getLog(Analysis.class);
	private static final long MAX_THREAD_WAIT_TIME_MS = 1000;
	private final ClassAndMethodContainer classAndMethodContainer;
	private final MIProject modelProject;
	private final Object analysisController;
	private final Object analysisControllerThread;
	private final UpdateDisplaysThread updateDisplaysThread;

	public Analysis(final ClassLoader classLoader, final MIProject modelProject) throws AnalysisInstantiationException {
		try {
			this.modelProject = modelProject;
			this.classAndMethodContainer = new ClassAndMethodContainer(classLoader);

			final Method createMethod = this.classAndMethodContainer.getAnalysisControllerCreateAnalysisController();
			final Object controllerAndMapping = ClassAndMethodContainer.invokeClassMethod(createMethod, null, this.modelProject, classLoader);
			this.analysisController = ClassAndMethodContainer.invokeMethod(this.classAndMethodContainer.getAnalysisControllerWithMappingGetController(),
					controllerAndMapping, null);
			this.analysisControllerThread = ClassAndMethodContainer.invokeConstructor(this.classAndMethodContainer.getAnalysisControllerThreadConstructor(),
					this.analysisController);
			this.updateDisplaysThread = new UpdateDisplaysThread(ClassAndMethodContainer.invokeMethod(
					this.classAndMethodContainer.getAnalysisControllerWithMappingGetMapping(), controllerAndMapping, null));
			if ((this.analysisController == null) || (this.analysisControllerThread == null)) {
				throw new AnalysisInstantiationException("An error occured while instantiating the analysis.");
			}
		} catch (final ProjectLoadException ex) {
			throw new AnalysisInstantiationException("An error occured while instantiating the analysis.", ex);
		} catch (final NullPointerException ex) {
			throw new AnalysisInstantiationException("An error occured while instantiating the analysis.", ex);
		}
	}

	public void start() throws AnalysisAlreadyStartedException {
		synchronized (this) {
			try {
				this.classAndMethodContainer.getAnalysisControllerThreadStart().invoke(this.analysisControllerThread, new Object[0]);
				this.updateDisplaysThread.start();
			} catch (final IllegalThreadStateException ex) {
				throw new AnalysisAlreadyStartedException("The analysis has already been started once. It has to be reinitialized to be runnable again.", ex);
			} catch (final IllegalAccessException ex) {
				Analysis.LOG.error("An error occured during a reflection method call.", ex);
			} catch (final IllegalArgumentException ex) {
				Analysis.LOG.error("An error occured during a reflection method call.", ex);
			} catch (final InvocationTargetException ex) {
				throw new AnalysisAlreadyStartedException("The analysis has already been started once. It has to be reinitialized to be runnable again.", ex);
			}
		}

	}

	public void stop() {
		synchronized (this) {
			try {
				this.classAndMethodContainer.getAnalysisControllerThreadTerminate().invoke(this.analysisControllerThread, new Object[0]);
				this.updateDisplaysThread.terminate();

				this.classAndMethodContainer.getAnalysisControllerThreadJoin().invoke(this.analysisControllerThread, Analysis.MAX_THREAD_WAIT_TIME_MS);
				this.updateDisplaysThread.join(Analysis.MAX_THREAD_WAIT_TIME_MS);
			} catch (final InterruptedException ex) {
				// Log the exception, but don't handle it further
				Analysis.LOG.info("An interrupted exception occured.", ex);
			} catch (final IllegalAccessException ex) {
				Analysis.LOG.error("An error occured during a reflection method call.", ex);
			} catch (final IllegalArgumentException ex) {
				Analysis.LOG.error("An error occured during a reflection method call.", ex);
			} catch (final InvocationTargetException ex) {
				Analysis.LOG.error("An error occured during a reflection method call.", ex);
			}
		}
	}

	public Enum<?> getCurrentState() {
		synchronized (this) {
			return (Enum<?>) ClassAndMethodContainer.invokeMethod(this.classAndMethodContainer.getAnalysisControllerGetState(), this.analysisController, null);
		}
	}

	public Object getDisplay(final String viewName, final String displayName) {
		return this.updateDisplaysThread.getDisplay(viewName, displayName);
	}

	/**
	 * This manager is responsible for the currently used and running instances of {@code AnalysisController}. It supplies methods to check the states of analysis,
	 * instantiate, start and to stop them.
	 * 
	 * @author Nils Christian Ehmke
	 * @version 1.0
	 */
	private class UpdateDisplaysThread extends Thread {

		/**
		 * This is the time the thread waits between the updates.
		 */
		private static final long SLEEP_TIME_MS = 2 * 1000;
		private final Object pluginMap;
		/**
		 * The map containing all available display objects.
		 */
		private final Map<String, Map<String, Object>> displayObjects = new ConcurrentHashMap<String, Map<String, Object>>(); // NOPMD
		/**
		 * This map contains the mapping to get the methods to be called.
		 */
		private final Map<Object, Method> methodMap = new ConcurrentHashMap<Object, Method>(); // NOPMD (Use of concurrent hash map)
		/**
		 * The field determining whether the thread has been terminated or not. // (USeConcurrentHashMap)
		 */
		private volatile boolean terminated = false;

		public UpdateDisplaysThread(final Object object) {
			this.pluginMap = object;
		}

		@PostConstruct
		private void initialize() {
			// Initialize the hashmaps and the necessary objects
			for (final MIView view : Analysis.this.modelProject.getViews()) {
				final Map<String, Object> viewMap = new ConcurrentHashMap<String, Object>(); // NOPMD (Use of concurrent hash map)
				this.displayObjects.put(view.getName(), viewMap);
				for (final MIDisplayConnector displayConnector : view.getDisplayConnectors()) {
					// final Method displayMethod = this.getDisplayMethod(this.pluginMap.get(displayConnector.getDisplay().getParent()).getClass(),
					// displayConnector.getDisplay().getName());

					// Make sure that the method really exists and that is has the correct parameters
					/*
					 * if ((displayMethod != null) && (displayMethod.getParameterTypes().length == 1)) {
					 * final Class<?> parameterType = displayMethod.getParameterTypes()[0];
					 * final Object displayObject;
					 * // Get the correct type
					 * if (Analysis.this.classAndMethodContainer.getImageClass().isAssignableFrom(parameterType)) {
					 * displayObject = new Image();
					 * } else {
					 * if (Analysis.this.classAndMethodContainer.getPlainTextClass().isAssignableFrom(parameterType)) {
					 * displayObject = new PlainText();
					 * } else {
					 * if (Analysis.this.classAndMethodContainer.getHtmlTextClass().isAssignableFrom(parameterType)) {
					 * displayObject = new HtmlText();
					 * } else {
					 * displayObject = null;
					 * }
					 * }
					 * }
					 * if (displayObject != null) {
					 * viewMap.put(displayConnector.getName(), displayObject);
					 * this.methodMap.put(displayObject, displayMethod);
					 * }
					 * }
					 */
				}
			}
		}

		/**
		 * Delivers the correct method to be invoked for an update of the display.
		 * 
		 * @param clazz
		 *            The class of the plugin.
		 * @param displayName
		 *            The name within the display-annotation.
		 * @return The method instance.
		 */
		private Method getDisplayMethod(final Class<? extends AbstractPlugin> clazz, final String displayName) {
			final Method[] methods = clazz.getMethods();
			for (final Method method : methods) {
				try {
					final Annotation displayAnnot = method.getAnnotation(Analysis.this.classAndMethodContainer.getDisplayAnnotationClass());
					if ((displayAnnot != null)
							&& (Analysis.this.classAndMethodContainer.getDisplayNameMethod().invoke(displayAnnot, new Object[0]).equals(displayName))) {
						// We found the correct method
						return method;
					}
				} catch (final IllegalAccessException ex) {
					// Log this exception, but ignore it
					Analysis.LOG.info("An error occured while calling a method.", ex);
				} catch (final IllegalArgumentException ex) {
					// Log this exception, but ignore it
					Analysis.LOG.info("An error occured while calling a method.", ex);
				} catch (final InvocationTargetException ex) {
					// Log this exception, but ignore it
					Analysis.LOG.info("An error occured while calling a method.", ex);
				}
			}
			return null;
		}

		/**
		 * Delivers the display object for the given view and the given display.
		 * 
		 * @param viewName
		 *            The name of the view.
		 * @param displayName
		 *            The name of the display in the view.
		 * @return The object for the given display if it exists, null otherwise.
		 */
		public Object getDisplay(final String viewName, final String displayName) {
			return this.displayObjects.get(viewName).get(displayName);
		}

		@Override
		public void run() {
			// Run until we have been interrupted
			while (!this.terminated) {
				for (final MIView view : Analysis.this.modelProject.getViews()) {
					this.displayObjects.get(view.getName());
					for (final MIDisplayConnector displayConnector : view.getDisplayConnectors()) {
						// final Object displayObject = viewMap.get(displayConnector.getName());
						// final Object pluginObject = this.pluginMap.get(displayConnector.getDisplay().getParent());
						// Update the display object
						// try {
						// this.methodMap.get(displayObject).invoke(pluginObject, displayObject);
						// } catch (final IllegalAccessException ex) {
						// Analysis.LOG.error("An error occured while updating the display.", ex);
						// } catch (final IllegalArgumentException ex) {
						// Analysis.LOG.error("An error occured while updating the display.", ex);
						// } catch (final InvocationTargetException ex) {
						// Analysis.LOG.error("An error occured while updating the display.", ex);
						// } catch (final NullPointerException ex) { // NOPMD
						// Analysis.LOG.error("An error occured while updating the display.", ex);
						// }
					}
				}

				// Wait a little bit.
				try {
					Thread.sleep(UpdateDisplaysThread.SLEEP_TIME_MS);
				} catch (final InterruptedException ex) {
					// We have been interrupted. Exit the thread
					return;
				}
			}

		}

		public void terminate() {
			this.terminated = true;
		}
	}
}
